<?php

// settings
ini_set("display_errors", 1);// debug
ini_set('error_reporting', E_ALL);// debug
$ftp_server = "ftp.example.org";
$ftp_user = "user";
$ftp_pass = "password";
$ftp_dir = "htdocs";// use "." to copy the root directory or "dir/sub" (without ending slash) to copy a subdirectory
$local_dir = ".";// use "." to copy into the root directory or "dir/sub" (without ending slash) into a subdirectory
$ftp_passive_mode = true;// solves several problems
$files_per_cycle = 300;
$file_check_size = false;
$file_check_time = true;

// functions
function ftpexit($error_msg="") {
	global $ftp;
	ftp_close($ftp);
	exit($error_msg . "<br />\n");
}
function _get($var, $else=null, $force=false) {
	if (!$force) {
		return isset($_GET[$var]) ? $_GET[$var] : $else;
	}
	else {
		$force = array_flip($force);
		return isset($_GET[$var]) && isset($force[$_GET[$var]]) ? $_GET[$var] : $else;
	}
}
function permstr_dec($str) {
	$str = strtr(str_pad($str, 9, '-'), array(
		'-' => '0',
		'r' => '4',
		'w' => '2',
		'x' => '1'
	));
	return octdec(($str[0] + $str[1] + $str[2]) . ($str[3] + $str[4] + $str[5]) . ($str[6] + $str[7] + $str[8]));
}
function format_size($value) {
	$units = array(1099511627776 => 'TiB', 1073741824 => 'GiB', 1048576 => 'MiB', 1024 => 'KiB', 1 => 'Bytes');
	foreach ($units as $step => $unit) {
		if ($value >= $step) {
			return round($value / $step, 2) . ' ' . $unit;
		}
	}
	return 0;
}
function unix_file_type($str) {// http://en.wikipedia.org/wiki/Unix_file_types
	switch ($str) {
		case "-":
		case "r":
		case "w":
		case "x":
			return "file";
		case "d":
			return "dir";
		case "l":
			return "link";
		case "p":
			return "fifo";// named pipe
		case "s":
			return "socket";
		case "c":
			return "char";// device file for character devices
		case "b":
			return "block";// device file for block devices
	}
	return "unknown";
}
function rmdirs($dir) {
	$files = glob($dir . '*');
	if ($files) {
		foreach ($files as $file) {
			if (is_dir($file)) {
				rmdirs($file . '/');
			}
			else {
				if (!@unlink($file)) {
					chmod($file, 0777);
					unlink($file);
				}
			}
		}
	}
	if (!@rmdir($dir)) {
		chmod($dir, 0777);
		rmdir($dir);
	}
}
//rmdirs("example_folder/");exit;// debug
function js_redirect($url, $timeout=100) {
	ftpexit("<script>window.setTimeout('location.replace(\"" . $url . "\");', $timeout);</script>");
}
function ftp_sync($ftp_dir, $local_dir, &$last_file) {
	global $ftp;
	global $file_check_size, $file_check_time;
	global $files_per_cycle, $i;
	global $start_time, $files_count, $files_size, $update_count, $insert_count;
	$new_last_file = "";
	//echo 'ftp_rawlist: ' . $ftp_dir . "<br />\n";
	$files = ftp_rawlist($ftp, "-A " . $ftp_dir);
	foreach ($files as $file) {
		list($perms, $links, $user, $group, $size, $d1, $d2, $d3, $name) = preg_split('/\s+/', $file, 9);
		$time = strtotime(implode(' ', array($d1, $d2, $d3)));
		$type = unix_file_type($perms[0]);
		$name = "/" . $name;
		$perms = permstr_dec(substr($perms, 1));
		// skip until last written file has been reached
		if ($last_file) {
			if (strpos($last_file, $local_dir . $name) === false) {
				continue;
			}
			// last file found, skip to next file
			else if ($last_file == $local_dir . $name) {
				$last_file = "";
				continue;
			}
		}
		// write files
		// todo: add ftp login from local server to raise permissions in creating files
		switch ($type) {
			case "dir":
				if (!file_exists($local_dir . $name)) {
					if (!mkdir($local_dir . $name, $perms, true)) {
						ftpexit("Couldn't create dir " . $local_dir . $name . " ($perms)");
					}
					else if (!touch($local_dir . $name, $time)) {
						ftpexit("Couldn't change time " . $local_dir . $name . " ($time)");
					}
					//echo $local_dir . $name . " created<br />\n";
				}
				// recursive
				ftp_sync($ftp_dir . $name, $local_dir . $name, $last_file);
				break;
			case "file":
				$i++;
				$files_count++;
				$create_reason = false;
				if (!file_exists($local_dir . $name)) {
					$create_reason = ' (did not exist)';
					$insert_count++;
				}
				else if ($file_check_size && filesize($local_dir . $name) != $size) {
					$create_reason = ' (size differed)';
					$update_count++;
				}
				else if ($file_check_time && filemtime($local_dir . $name) != $time) {
					$create_reason = ' (source was newer)';
					$update_count++;
				}
				if ($create_reason) {
					$fh = fopen($local_dir . $name, 'wb');
					if (!ftp_fget($ftp, $fh, $ftp_dir . $name, FTP_BINARY)) {
						fclose($fh);
						unlink($local_dir . $name);
						ftpexit("There was a problem while downloading " . $ftp_dir . $name . " to " . $local_dir . $name);
					}
					else if (!touch($local_dir . $name, $time)) {
						fclose($fh);
						unlink($local_dir . $name);
						ftpexit("Couldn't change time " . $local_dir . $name . " ($time)");
					}
					fclose($fh);
					chmod($local_dir . $name, $perms);
					$files_size += $size;
					//echo $local_dir . $name . " $create_reason<br />\n";
				}
				$new_last_file = $local_dir . $name;
				break;
			default:
				ftpexit("The file $name has an unsupported filetype $type");
		}
		if ($i >= $files_per_cycle) {
			break;
		}
	}
	if ($i >= $files_per_cycle && $new_last_file) {
		echo 'Processed: ' . $files_count . "<br />\n";
		echo 'Transferred: ' . ($files_size ? format_size($files_size) : 0) . "<br />\n";
		echo 'Transfer Rate: ' . ($files_size ? format_size(round($files_size / (time() - $start_time))) : "0 KiB") . "/s<br />\n";
		echo 'Updated: ' . $update_count . "<br />\n";
		echo 'Inserted: ' . $insert_count . "<br />\n";
		echo 'Last File: ' . $new_last_file . "<br />\n";
		echo 'Refresh Time: ' . date('d.m.Y H:i:s') . "<br />\n";
		js_redirect($_SERVER['SCRIPT_NAME'] . '?lf=' . urlencode($new_last_file) 
																				. '&st=' . $start_time
																				. '&fs=' . $files_size
																				. '&uc=' . $update_count
																				. '&ic=' . $insert_count
																				. '&fc=' . $files_count
								);
	}
}
// set vars
$i = 0;
$start_time = intval(_get("st", time()));
$files_count = intval(_get("fc", 0));
$files_size = intval(_get("fs", 0));
$update_count = intval(_get("uc", 0));
$insert_count = intval(_get("ic", 0));
$last_file = _get("lf");

echo "<b>FTP-Syncy</b><br />\n";
echo 'Start Time: ' . date('d.m.Y, H:i:s', $start_time) . "<br />\n";
echo 'Start File: ' . htmlspecialchars($last_file) . "<br />\n";

// start ftp sync
$ftp = ftp_connect($ftp_server) or die("Couldn't connect to $ftp_server");
if (!ftp_login($ftp, $ftp_user, $ftp_pass)) {
	ftpexit("Couldn't connect as $ftp_user");
}
ftp_pasv($ftp, $ftp_passive_mode);
ftp_sync($ftp_dir, $local_dir, $last_file);

echo 'Processed: ' . $files_count . "<br />\n";
echo 'Transferred: ' . ($files_size ? format_size($files_size) : 0) . "<br />\n";
echo 'Transfer Rate: ' . ($files_size ? format_size(round($files_size / (time() - $start_time))) : "0 KiB") . "/s<br />\n";
echo 'Updated: ' . $update_count . "<br />\n";
echo 'Created: ' . $insert_count . "<br />\n";
echo 'Finish Time: ' . date('d.m.Y, H:i:s') . "<br />\n";
ftpexit();
?>